#!/usr/bin/perl

use 5.006;
use strict;
use warnings;
use Carp;

use POSIX qw(strftime setsid floor);
##use Net::Ping;
##use Net::Server;
##use vars qw(@ISA $VERSION);

##@ISA = qw(Net::Server);
##$VERSION = '1.0';

use IPC::Run qw(start finish run pump);
use AppConfig qw/:argcount :expand/;
use IO::File;
use IO::All qw/io/;
use EV;

##use NetServer::Portal;
##NetServer::Portal->default_start();
##warn "NetServer::Portal listening on port ".(7000+($$ % 1000))."\n";

my $config = AppConfig->new(
    "host|h=s@"         => {DEFAULT => []},
    "port|p=s"          => {DEFAULT => 8999},
    "corpus=f@"         => {DEFAULT => []},
    "exclude=s@"        => {DEFAULT => []}, ## exclude corpus
    "parser=s"          => {DEFAULT => "frmgtel"},
    "rparser=s"         => {DEFAULT => "frmgtelr"},
    "input|in|i|t=s"    => {DEFAULT => ''},
    "verbose|v!"        => {DEFAULT => 1},
    "stats|s!"          => {DEFAULT => 0},

    "forest!"           => {DEFAULT => 0},
    "display|d=s"       => {DEFAULT => ''},
#    "dependency!"       => {DEFAULT => 0}, # display graph
    "grammar!"          => {DEFAULT => 0},
    "tagger!"           => {DEFAULT => 0},
    "xmldep!"           => {DEFAULT => 0},
    "lpdep!"            => {DEFAULT => 0},
    "dotdep!"           => {DEFAULT => 0},
    "easy!"             => {DEFAULT => 0},
    "easyhtml!"         => {DEFAULT => 0},

    "errfile=s"         => {DEFAULT => "errors"},
    "setsid!"           => {DEFAULT => 0},
    "log_file=f"        => {DEFAULT => '/tmp/dispatch.log'},
    "status!"           => {DEFAULT => 0},
    "time!"             => {DEFAULT => 0},
    "robust!"           => {DEFAULT => 0},
    "easyinput"         => {DEFAULT => 0},
    "timeout=i"         => {DEFAULT => 200},

    "colldir=f"         => {DEFAULT => "./results"},
    "collsave!"         => {DEFAULT => 0},
    "collbase=f"        => {DEFAULT => 'sentence'},,

    "allparse!"         => {DEFAULT => 0},
    "showdag!"          => {DEFAULT => 0},
    "run!"              => {DEFAULT => 1},
    "final!"            => {DEFAULT => 1},
    "disambiguate|dis!" => {DEFAULT => 0},
    "date!"             => {DEFAULT => 0},
    "printhost!"        => {DEFAULT => 0},
    "results=f"         => {DEFAULT => 'results'},
    "dagdir=f"          => {DEFAULT => 'dag'},
    "dagext=s"          => {DEFAULT => "udag"},
    "lock=f"            => {DEFAULT => 'lock'},

    "compress!"         => {DEFAULT => 0},
    "kill=i"            => {DEFAULT => 0},
			    "adj_max=i" => {DEFAULT => 6},
    );

my $conffile = 'dispatch.conf';

if (@ARGV && $ARGV[0] =~ m/^--config/) {
  shift @ARGV;
  $conffile = shift @ARGV;
  print STDERR "USING config file $conffile\n";
}

if (-r $conffile) {
  $config->file("$conffile")
    or die "can't open or process configuration file $conffile, stopped";
}

$config->args();

## retrieve hosts
my @hosts = @{$config->host()};
our $dispatch;

#excute callparser on every host
if (@hosts <= 1 && !-r $conffile) {
    my $client = MyClient->new($config, $hosts[0], $config->port, 1);
    $client->set_collbase($config->collbase);
    $client->set_colldir($config->colldir);
    $client->process_input;
    EV::loop;
}
else {
    my %corpora = ();
    print "Starting Dispatcher\n";
    $dispatch = MyDispatcher->new($config);
}


############################################
############################################

package MyDispatcher;
use Carp;
use EV;
use IO::Socket;
use IO::All qw/io/;
use POSIX qw(strftime);
use File::Basename;
use NDBM_File; 


sub DESTROY {
  my $self = shift;
  ##  print "DESTROY!\n";
  my $pending = $self->{pendingfile};
  open(PENDING,">$pending") or die "can't open $pending: $!";
  foreach my $x (values %{$self->{processing}}) {
    print PENDING "$x->{corpus}\t$x->{sid}\n";
  }
  close(PENDING);
  $self->verbose("Quitting");
}

############################################
# Purpose    : Create and launch dispatcher
#
sub new {
    my $this   = shift;
    my $class  = ref($this) || $this;
    my $config = shift;

    print "HOST @hosts\n";

    my $self = { config      => $config,
		 hosts       => {},
		 corpus      => {}, # corpus and info. not used
		 lcorpus     => [], # corpus list
		 tasks       => [],
		 ##		 pending  => {}, # hash (host => task) waiting to be finished
		 ctask       => 0,
		 pendingfile => $config->results."/status/pending.log",
		 sentences   => $config->results."/sentences.log",
		 maxslog     => 10_000, # max lines in sentences.log
		 n           => 0, # current line in sentences.log
		 stop        => 0, # 1 when no more dags
		 dags        => [], # a queue of DAGs
		 dagsnb      => 0,  # the length of the DAG queue
	       };

    bless $self, $class;

    { no strict 'refs';
      ## the following to affect the adapted version of verbose to the dispatcher !
      ## should then be faster
      ## moreover, it avoids some strange problems when terminating the program
      ## with $config being destroyed before the dispatcher
      my $code;
      if ($config->verbose) {
	$code = sub {
	  my $self = shift;
	  my $date = strftime "[%F %H:%M:%S]", localtime;
	  print "$date @_\n";
	  print VERBOSE "$date @_\n" if (defined $self->{log});
	};
      } else {
	$code = sub {};
      }
      *{'verbose'} = $code;
    }

    ## Preparation
    my $results   = $self->results;
    my $lock      = $self->lock; # used ?
    my $log       = "$results/dispatch.log";

    $self->verbose("Preparing results directory '$results'");
    mkdir $results;
    mkdir "$results/$lock";
    mkdir "$results/$lock.allow";
    mkdir "$results/status";

    system("cp $conffile $results/") if (-r $conffile);
    
    my $dagdir = $config->dagdir;
    my $dagext = $config->dagext;
    my %info   = ();
    
    if (!-r "$dagdir/info.txt") {

      ## retrieve dag info (corpus name and number of sentences for stats)
      foreach my $file (<$dagdir/*.$dagext>) {
	
	unless (open(FILE,"<$file")) {
	  warn "can't open corpus dag $file: $!";
	  next;
	}
	my $n = 0;
	my $corpus = basename($file,".$dagext");
	## not used for now !
	
	unless (-r "$dagdir/info.txt") {
	  
	  while (<FILE>) {
	    if (/^##\s*DAG\s+BEGIN/o) {
	      my ($i) = (<FILE> =~ /id="E(\d+)F\d+"/);
	      $n++;
	    }
	  }
	}
	
	close(FILE);
	
	$info{$corpus} = $n;
	
	# ?
	#$self->{corpus}{$corpus} = { min => 0, max => 0, size => $n };
	$self->{corpus}{$corpus} = { min => 0, max => 0};
      }
      
      $self->verbose("create info.txt");
      open(INFO,">>$results/info.txt") or die "can't open file $results/info.txt: $!, stopped";
      
      foreach my $corpus (sort {$info{$b} <=> $info{$a}} keys %info) {
	printf INFO "%-20s %4i\n",$corpus,$info{$corpus};
      }
      
      close(INFO);
    } else {
      system("cp $dagdir/info.txt $results/info.txt");
      open(INFO,"<$dagdir/info.txt") 
	or die "can't open $dagdir/info.txt ($!)";
      while(<INFO>) {
	next if (/^\#/);
	my ($corpus,$size) = split(/\s+/,$_);
	$info{$corpus} = $size;
	$self->{corpus}{$corpus} = { min => 0, max => 0};
      }
      close(INFO);
    }
    
    unless (@{$config->get('corpus')}) {
	$config->corpus($_) foreach (sort keys %{$self->{corpus}});
    }
    
    $self->{lcorpus} = $config->get('corpus');
    if (@{$config->exclude()}) {
	my (@lcorpus, $skip);
	foreach my $corpus (@{$self->{lcorpus}}) {
	    $skip = 0;
	    foreach my $excluded (@{$config->exclude()}) {
		$skip = 1 if ($corpus eq $excluded);
	    }
	    push (@lcorpus, $corpus) unless ($skip == 1);
	}
	@{$self->{lcorpus}} = @lcorpus;
    }
    $self->verbose("Corpus list : @{$self->{lcorpus}}");
    
    ## Open log files
    if ($config->get('verbose')) {
	open(VERBOSE,">>$log") or die "can't open log file $log: $!, stopped";
	VERBOSE->autoflush(1);
	$self->verbose("Activate log in $log");
	$self->{log} = $log;
	my $sentences = $self->{sentences};
	open(SENTENCES,">>$sentences") or die "can't open sentences log $sentences: $!, stopped";
	## SENTENCES->autoflush(1);
    }

    #$self->verbose("PID $$");


    ## Reading pending sentences from previous run
    my $pending = "$results/status/pending.log";
    if (-r $pending) {
      $self->verbose("Resuming processing");
      $self->{pending} = {};
      open(PENDING,"<$pending")  or die "can't open $pending: $!";
      while(<PENDING>){
	my ($corpus,$sid) = split(/\s+/,$_);
	$self->verbose("Registering pending $corpus $sid");
	$self->{pending}{$corpus}{$sid} = 1;
      }
      close(PENDING);
    }

    ## trap deaths in callbacks
    $EV::DIED = sub {
      my $error = $@;
      $self->verbose('Event DIED');
      if (ref($error) eq "ErrorConnection") {
	## intercept connection errors and reconnect
	$error->{client}->restart_connection;
      } else {
	$self->verbose("DIED: could not handle '$error'");
	$error->{client}->restart_connection;
      }
    };


    # kill dispatcher after x hours (for grid'5000)
    my $kill = $config->get('kill');
    unless ($kill eq '0') {
	EV::timer($kill*3600,
		  0,
		  sub {
		    $self->verbose("kill after $kill hours!");
		    kill 2, $$;
		  },
		 );
      }

    # delete ghost hosts
#     Event->var(var  => \$self->{stop},
# 	       poll => 'w',
# 	       desc => 'Watching stop var',
# 	       cb   => sub {
# 		   foreach my $h ( keys(%{$self->{hosts}}) ){
# 		       my $client = $self->{hosts}{$h}{client};
# 		       if ($client->{timer}) {
# 			   $self->verbose("delete ghost $h");
# 			   $self->close_host($client);
# 		       }
# 		   }
# 	       });

    $self->{watchers} = [];

    push @{$self->{watchers}},
      EV::signal('INT',
		 sub { $self->verbose("Got signal INT!");
		       $self->close_all_hosts;
		       EV::unloop(EV::UNLOOP_ALL);
		       $self->exit;
		     });
    
    push @{$self->{watchers}},
      EV::signal( 'PIPE', 
		  sub {   ## do no exit on PIPE
		    $self->verbose("Got signal PIPE");
		  }
		);

    push @{$self->{watchers}},
      EV::signal('HUP', sub { print("GOT SIGHUP!\n") });
    
    push @{$self->{watchers}},
      EV::signal('QUIT', 
		 sub { print("GOT SIGQUIT!\n");
		       $self->close_all_hosts;
		       EV::unloop(EV::UNLOOP_ALL);
		       $self->exit;
		     });

    push @{$self->{watchers}},
      EV::signal('ILL', 
		 sub { print("GOT SIGILL!\n");
		       $self->close_all_hosts;
		       EV::unloop(EV::UNLOOP_ALL);
		       $self->exit;
		     });
    
    push @{$self->{watchers}},
      EV::signal('ABRT', 
		 sub { print("GOT SIGABRT!\n");
		       $self->close_all_hosts;
		       EV::unloop(EV::UNLOOP_ALL);
		       $self->exit;
		     });

    push @{$self->{watchers}},
      EV::signal('TERM', 
		 sub { print("GOT SIGTERM!\n");
		       $self->verbose("GOT SIGTERM!");
		       $self->close_all_hosts;
		       EV::unloop(EV::UNLOOP_ALL);
		       $self->exit;
		     });

     $self->{dag_watch} =
       EV::idle(
 	       sub { 
 		 if ($self->{dagsnb} < 1000) {
		   ## $self->verbose("Idle $self->{dagsnb}");
		   $self->push_next_dag;
		 } else {
		   $self->{dag_watch}->stop;
		 }
 	       }
 	      );
    $self->{dag_watch}->priority(EV::MINPRI);

    my @hosts = @{$self->{config}->host};
    # register callparser on every host
    foreach my $host (@hosts) {
	$self->add_host($host);
    }


##    EV::set_timeout_collect_interval 0.03;
    EV::loop;
    ##    $self->exit;
}

sub try_corpus {
  my ($self,$corpus) = @_;
  ## a corpus should be tried,
  ## either because it is a pending one or because there is no longer pending corpora
  return (! (exists $self->{pending})) || (exists $self->{pending}{$corpus});
}

sub try_dag {
  my ($self,$corpus,$sid) = @_;
  ##  print "Try dag: $corpus $sid\n";
  return 1 unless (exists $self->{pending});
  return 0 unless ($self->{pending}{$corpus}{$sid});
  $self->verbose("Resuming phase: selecting dag $corpus $sid");
  delete $self->{pending}{$corpus}{$sid};
  delete $self->{pending}{$corpus} unless (keys %{$self->{pending}{$corpus}});
  delete $self->{pending} unless (keys %{$self->{pending}});
  return 1;
}

############################################
# Purpose    : create client object for a host
# Parameters : $host
#
sub add_host {
    my ($self, $host) = @_;
    $self->verbose("adding $host");

    my $port;
    if ($host =~ /(\S+):(\S+)/) {
	$host = $1;
	$port = $2;
    }

    my $hostlabel = $host;
    if ($host =~ /(\S+)\@(\S+)/) {
      $host = $2;
      $hostlabel = $1;
    }

    if (exists $self->{hosts}{$hostlabel}{client}) {
	$self->verbose("host '$host' already registered");
	return;
    }

    # register parserd client
    my $client = MyClient->new($config, $host, $port, 0, $self, $hostlabel);
    $self->{hosts}{$hostlabel}{client} = $client;
}

sub close_all_hosts {
  my $self = shift;
  $self->verbose("Closing all hosts");
  foreach my $info (values %{$self->{hosts}}) {
    next unless (defined $info);
    my $client = $info->{client};
    next unless (defined $client);
    $self->verbose("Closing host '$client->{hostlabel}'");
    $client->closing;
    undef $info;
  }
}

sub results {
    my $self = shift;
    return $self->{config}->results;
}

sub dagdir {
    my $self = shift;
    return $self->{config}->dagdir;
}

sub dagext {
    my $self = shift;
    return $self->{config}->dagext;
}


sub lock {
    my $self = shift;
    return $self->{config}->lock;
}

sub warning {
    my $self = shift;
    print STDERR @_,"\n";
}

# Some commands to communicate with the server of parsers 
# through telnet during corpus processing
sub user_handler {
    my ($self, $child, $info) = @_;

    return if ($info =~ /^\s*$/);

    if ($info =~ /^quit|exit/){
	$child->print("Good bye");
	$child->close;
    } elsif ($info =~ /^hosts/) {
	$child->print("hosts ",join(" ",sort(keys %{$self->{hosts}})),"\n");
    } elsif ($info =~ /^add\s+(\w+)/) {
	$self->add_host($1);
	$child->print("done\n");
    } elsif ($info =~ /^del\s+(\w+)/) {
    } elsif ($info =~ /^corpus/) {
	foreach my $corpus (sort keys %{$self->{corpus}}) {
	    $child->print("$corpus $self->{corpus}{$corpus}{size}\n");
	}
    } else {
	$child->print("Bad user cmd: $info\n");
	$self->out("Bad user cmd: $info");
    }

}

sub exit {
    my $self=shift;
    CORE::exit;
}

sub out {
    my $self = shift;
    autoflush STDOUT 1;
    print join(' ',@_),"\n";
}

sub close_host {
    my ($self, $client) = @_;
    my $hostlabel = $client->{hostlabel};
    $self->verbose("Closing host '$hostlabel'");
    $client->closing;
    delete $self->{hosts}{$hostlabel};
    my @h = keys %{$self->{hosts}};
    ##    $self->verbose("... remaining hosts: @h");
    unless (keys %{$self->{hosts}}) {
	my $done = $config->results."/status/done";
	system("touch $done");
	$self->{watchers} = [];
	EV::unloop(EV::UNLOOP_ALL);
    }
}

############################################
# Purpose    : Pop dag from corpus and send it to server of parsers
# Parameters : $child, $info
#

sub push_next_dag {
  my $self = shift;

  ## unless corpus already open
  unless (exists $self->{handle}) {
  START:
    ## if no remaining corpus to process, then close host
    unless (@{$self->{lcorpus}}) {
      $self->{stop} ||= 1;
      $self->verbose("Deleting dag_watcher");
      undef $self->{dag_watch};
      return;
    }

    # new corpus to process
    my $results = $self->results;
    my $dagdir  = $self->dagdir;
    my $dagext  = $self->dagext;
    my $corpus  = shift @{$self->{lcorpus}};
    my $sid     = 0;

    unless ($self->try_corpus($corpus)) {
      ##	  $self->verbose("Resuming phase: shifting $corpus");
      goto START;
    }
    if (-r "$results/status/$corpus") {
      my $status = IO::All::io("$results/status/$corpus")->[0];
      chomp $status;
      if ($status eq 'one') {
	$self->verbose("Shifting $corpus");
	goto START;
      }
      $sid = $status;
      $self->verbose("Resuming $corpus at '$sid'");
    }
    $self->verbose("Opening $corpus");
    my $file = "$dagdir/$corpus.$dagext";
    if (-r $file) {
      $self->{handle} = new IO::File "<$file"
	or die "can't open dag file $file, stopped ($!)";
      $self->{handle}->autoflush(1);
    } elsif (-r "$file.bz2") {
      my $h = IO::Handle->new;
      open($h,"bzcat $file.bz2 |") 
	or die "can't open dag file $file, stopped ($!)";
      $self->{handle} = $h;
    } else {
      $self->verbose("*** Dag file $file not found");
      goto START;
    }
    mkdir "$results/$corpus";
    $self->{current} = $corpus; # store current corpus name
    my $corpuslog = "$results/$corpus.log";
    my $io = $self->{currentlog} = io("$corpuslog")
      || die "can't open new corpus log '$corpuslog':$!";
    $io->mode('>>')->open;
    if ($sid) {
      my $stop = 0;
      my $handle = $self->{handle};
      while (<$handle>) {
	if (/^##\s*DAG\s+END/o && $stop) {
	  $self->verbose("Found resuming $corpus at '$sid'");
	  last;
	}
	if (/^##\s*DAG\s+BEGIN/o) {
	  $_ = <$handle>;
	  my ($xsid) = /E(\d+(?:\.\d+)?)F\d+/;
	  $stop=1 if  ($xsid eq $sid);
	  next;
	}
      }
    }
  } ## end unless (exists $self->{handle})
  
 START2:
  ## if current corpus finished
  if (eof($self->{handle})) {
    $self->verbose("Closing corpus '$self->{current}'");
    $self->{handle}->close;
    delete $self->{handle};
    goto START;
  }
  
  my $corpus = $self->{current};
  

  ## The corpus is now open. Ready to pop next dag
  
  my $sid; # sentence id
  my @dag;
  
  my $handle = $self->{handle};
  while (<$handle>) {
    unless (defined $sid) {
      ($sid) = /id="E(\d+(?:\.\d+)?)F\d+"/;
    }
    push(@dag,$_);
    last if (/^##\s*DAG\s+END/o);
  }
  
  unless ($self->try_dag($corpus,$sid)) {
    ##      $self->verbose("Resuming phase: shifting dag $corpus $sid");
    $sid = undef;		# dont forget to reset sid and @dag !
    @dag=();
    goto START2;
  }

##  $self->verbose("Pushing sentence $corpus $sid");
  push(@{$self->{dags}},
       { dag => [@dag],
	 corpus => $corpus,
	 sid => $sid,
	 corpuslog => $self->{currentlog}
       }
      );
  my $n = $self->{dagsnb}++;
  if ($n < 200 && !$self->{dag_watch}->is_active) {
    $self->{dag_watch}->start;
  }
}

sub next_dag {
    my ($self, $client) = @_;

    #return $self->user_handler($child,$info) unless ($info =~ /^next\s+(\S+)/);

    while ($self->{dagsnb}==0) {
      if ($self->{stop} == 1) {
	$self->close_host($client);
	return;
      }
      $self->push_next_dag;
    }

    my $dag = shift @{$self->{dags}};
    $self->{dagsnb}--;
    my $corpus = $dag->{corpus};
    my $sid = $dag->{sid};
##    $self->verbose("Selecting sentence $corpus $sid");

    my $host = $client->{hostlabel};
    $client->set_init_state; #change client state: wait for server answer

    ## $self->verbose("get request from $host");
    my $hostinfo = $self->{hosts}{$host};

    $hostinfo->{requests}++;

    ## Check if change of corpus since last dag (or if first time)
    if (!defined $self->{hosts}{$host}{current} # first time for host
	|| $corpus ne $self->{hosts}{$host}{current}
	) {
      my $old = $self->{hosts}{$host}{current};
      $self->{hosts}{$host}{current} = $corpus;
      $self->verbose("$host : selection of corpus $corpus");
      $client->{log} = $dag->{corpuslog};
      $client->corpus_start($corpus, $self->results."/$corpus");
      if (defined $old) {
	my @h = grep {defined $self->{hosts}{$_}{current} && ($self->{hosts}{$_}{current} eq $old)} keys %{$self->{hosts}};
	unless (@h) {
	  $self->verbose("Finished handling $old");
	  ## could add some hook at this point
	  ## for instance, to backup the results for $old
	}
      }
    }

    my $date = strftime "[%F %H:%M:%S]", localtime;
    print SENTENCES "$date $corpus $sid $host\n";
    $self->{n}++;

    if ($self->{n} == $self->{maxslog}) {
	close(SENTENCES)
	  or die "can't close sentence log: $!";
	my $sentences = $self->{sentences};
	if (-e "$sentences.1") {
	  rename ("$sentences.1", "$sentences.2");
	}
	rename ($sentences, "$sentences.1");
	$self->{n} = 0;
	open(SENTENCES,">$sentences") 
	  or die "can't open sentences log $sentences: $!, stopped";
    }

    #$self->verbose("Trying $corpus $sid on $host");

    $self->{processing}{$host} = { corpus => $corpus, sid => $sid };
    $client->process_input($corpus, @{$dag->{dag}});

}

1;

############################################
############################################

package MyClient;

use strict;
use warnings;

use IO::All qw/io/;
use IO::Socket;
use IO::Socket::INET;
use POSIX qw(strftime setsid floor);
use IPC::Run qw(start finish run pump);
use File::Temp qw/tempfile/;
use Time::HiRes qw(gettimeofday tv_interval);

my ($io, $report, @failed, %dagwords, %stats);


our $formats;

sub init {
  return if ($formats);
  $formats =
    {
     tagger   => { forest => 'tagger',
		   ext => 'tag.txt'
		 },
     forest   => { forest => 'xml',
		   ext => 'der.xml',
		   bzip2 => 1
		 },
     xmldep   => { forest => 'xmldep',
		   ext => 'dep.xml',
		   bzip2 => 1
		 },
     lpdep    => { forest => 'lpdep' },
     dotdep   => { forest => 'dep',
		   ext => 'dot'
		 },
     stats    => { forest => 'stats' },
     easy     =>  { ext => 'easy.xml',
		    forest => 'easy',
		    bzip2 => 1
		  },
     grammar  => { forest => 'grammar',
		   ext => 'gr.html',
		   bzip2 => 1
		 },
     easyhtml => { forest => 'easyhtml',
		   ext => '.easy.html'
		 }
    };
}

if ($config->status) {
    use Term::Report;
    $report = Term::Report->new( fh => \*STDERR );
    $report->savePoint('total',   "Total: ",     0);
    $report->savePoint('success', "\nSuccess: ", 0);
    $report->savePoint('failed',  "\nFailed: ",  0);
}

############################################
# Purpose    : create client and process input
# Parameters : $config, $host, $port, $input, $final
# Comments   : if $final is true, then print final stats
#
sub new {

    my $self       = shift;
    my $config     = shift;
    my $host       = shift || 'localhost';
    my $port       = shift || '8999';
    my $final      = shift;
    my $dispatcher = shift;
    my $hostlabel = shift || $host;
    my (@handlers, @dag, @formats, @exts);

    init();
    
    %stats = 
	( tried    => 0,
	  skipped  => 0,
	  failed   => 0,
	  time     => 0,
	  clusters => 0,
	  nodes    => 0,
	  edges    => 0
	);

    my $class  = ref($self) || $self;

    my $log = io->stdout;
    
    $self = { #CONFIG   => $config, # config is in same file
	     dispatcher => $dispatcher,
	     host       => $host,
	     hostlabel => $hostlabel,
	     port       => $port,
	     input      => '',
	     output     => 'yesno', # expected output from server
	     client     => '0',
	     io         => \$io,
	     tmplog     => '',
	     log        => $log,
	     section    => "s0",
	     stats      => \%stats,
	     report     => \$report,
	     collection => '',
	     failed     => \@failed,
	     daglabel   => '',
	     dagwords   => \%dagwords,
	     handlers   => \@handlers,
	     formats    => \@formats,
	     exts       => \@exts,
	     final      => $final, # print final stats
	     dag        => \@dag,
	     outfile    => '',
	     collbase   => '',
	     colldir    => '',
	     seconds    => '5', # number of seconds before restarting connection
	     next      => 0,
	     states    => {}
	    };

    bless $self, $class;

    # create colldir
    mkdir $config->colldir if ($config->collsave && !-e $config->colldir);

    $config->set('display', 'dependency') if ($config->display eq 'dep'); # tmp

    $self->{parser} = ($config->robust) ? $config->rparser : $config->parser;

    my $last='yesno';
    $self->{states}{$last} = 
      { 
       handler =>  $self->build_yesno_handler
      };
    foreach my $f (keys %$formats) {
      next unless ($config->get($f));
      my $info = $formats->{$f};
      $self->{states}{$last}{next} = { format => $f,
				       forest => $info->{forest}
				     };
      $self->verbose("Register switch from $last to $f");
      $self->{states}{$f} = {  ext => $info->{ext},
			       handler => $self->get_handler($f),
			       next => 0
			    };
      $last = $f;
    }
    $self->start_connection;

    return $self;
}

sub verbose {
    my $self = shift;
    if ($self->{dispatcher}) {
	$self->{dispatcher}->verbose($self->{hostlabel}, ":", shift);
    } else {
	my $date = strftime "[%F %H:%M:%S]", localtime;
	print "$date $self->{hostlabel} : ", shift, "\n";
    }
}

############################################
# Purpose    : start telnet connection to host
# 
sub start_connection {
    my $self = shift;

    my $socket = IO::Socket::INET->new(Proto     => "tcp",
				       PeerAddr  => $self->{host},
				       PeerPort  => $self->{port});
    unless ($socket) {
	$self->verbose("cannot start connection, restart host in $self->{seconds} seconds");
	$self->{timer} = EV::timer($self->{seconds},
				   0,
				   sub {$self->start_connection;}
				  );
	$self->{seconds} *= 2;	# for next time !

      } else {

	$self->verbose("start connection");
	$self->{socket} = $socket;
	$self->{seconds} = 5;	# reinitialize
	##	$self->{created} = time; ## creation time of the client



	$self->send_directive_cmd("set mode data");
	$self->send_directive_cmd("set forest yesno");
	my $timeout = $config->get('timeout');
	$self->send_directive_cmd("timeout $timeout");
	$self->send_directive_cmd("limit cache 0"); # no sentence cache on server
	my $adj_max = $config->get('adj_max');
	$self->send_directive_cmd("set adj_max $adj_max"); # restrict ambiguities

	# (re)initialize watcher with new socket
	$self->{watcher} = EV::io($socket,
				  EV::READ,
				  $self->{states}{'yesno'}{handler}
				 );

	if (exists $self->{sentence_init}) {
	  ## if a failure were found before sending the initial command for a sentence
	  ## we keep the same sentence and retry
	  $self->verbose("Restarting $self->{label}");
	  $self->handle_sentence_aux($self->{sentence_init});
	} else {
	  ## otherwise we skip what remains to be done for the current sentence
	  ## and move to the next one
	  $self->flush_tmplog if ($self->{tmplog});
	  $self->{dispatcher}->next_dag($self) if ($self->{dispatcher});
	}
    }
}

## flush tmplog when finishing processing a sentence
## either in a normal way or because of some failure
## however, stopping dispatch.pl should not flush the log
## because it will build again when restarting
sub flush_tmplog {
  my $self = shift;
  if (my $tmplog = $self->{tmplog}) {
    $self->{log}->print($tmplog);
    $self->{tmplog} = '';
  }
}

sub restart_connection {
  my $self = shift;
  undef $self->{watcher};
  $self->start_connection;
}

sub raise_connection_error {
  my ($self, $msg) = @_;
  $self->verbose("$msg");
  $self->{socket}->close if ($self->{socket});
  delete $self->{socket};
  my $except = bless { client => $self, msg => $msg }, "ErrorConnection";
  die $except;
}

sub send_cmd {
    my ($self, $cmd) = @_;

    my $socket = $self->{socket};
    if (!$socket) {
      $self->raise_connection_error("pb1 print socket: reconnect (cmd=$cmd)");
    }
    
    unless (print $socket $cmd."\n") {
      $self->raise_connection_error("pb2 print socket: reconnect (cmd=$cmd)");
    }
}

sub send_directive_cmd {
  my ($self, $cmd) = @_;

  my $socket = $self->{socket};
  if (!$socket) {
    $self->raise_connection_error("pb1 print socket: reconnect (cmd=$cmd)");
  }

  unless (print $socket $cmd."\n") {
    $self->raise_connection_error("pb2 print socket: reconnect (cmd=$cmd)");
  }
  <$socket>;
}

sub get_output {
    my $self = shift;
    my $socket = $self->{socket};
    if (!$socket) {
	$self->verbose("socket undef");
	$self->raise_connection_error("pb read socket: reconnect");
	return 0;
    }
    my @output;
    while (<$socket>) {
      if (/^Bzip2:\s+length=(\d+)/){
	my $l = $1;
	my $buff='';
	$socket->read($buff,$l) || die "can't receive bzip2: $!";
	push(@output,$buff);
	next;
      }
      if (/Timed Out/) {
	push (@output, 'timeout');
	last;
      }
      if  (/<END_DATA>/) {
	if (/time=(\S+)/) {
	  $self->{cmd_time} = $1;
	}
	last;
      };
      chomp $_;
      push (@output, $_);
    }

    return @output;
}

############################################
# Purpose    : read input, decode dag, then calls handle_sentence
#              and finally ask for new dag
# Input      : can be @dag sent by the distpacher, or io input (sentence or dag)
#
sub process_input {
    my $self        = shift;
    $self->{corpus} = shift; # for record
    my @dag         = @_;

    my $io;
    my $next_line;

    if (@dag) {
      $next_line = sub {
	return undef unless @dag;
	my $x = shift @dag; 
##	$self->verbose("read dag $x");
	return $x;
      }
    } else {
      #$self->verbose("process dag");

      
      ## Reading sentences
      if ($config->get('input') eq '-') {
	$io = io->stdin;
      } 

      $next_line = sub {
	return $io->readline;
      }

    }

    my $label;
    my $dagmode    = 0;
    my @dagedges   = ();
    my $log        = $self->{log};

    while ($_=$next_line->()){

	#print "*$_\n";

	if (/^##\s*(\d+)(\..*)/) {
	    append $log "\n## Section $1$2\n\n" if ($config->verbose);
	    $self->{section} = "s$1";
	    next;
	}
	if (/^##\s*E(\d+)\.(\d+)$/) {
	    $label  = $1 - 1;
	    $label .= ".$2";
	    next;
	}
	if (/^##\s*E(\d+)$/) {
	    $label = $1 - 1;
	    next;
	}
	if (/^##\s*DAG\s+BEGIN/io) {
	    $dagmode  = 1;
	    @dagedges = ();
	    %{$self->{dagwords}} = ();
	    next;
	}

	## when dag ends, build sentence and decoded dag
	if ($dagmode && /^##\s*DAG\s+END/io) {
	    $dagmode     = 0;
	    my $sentence = '';
	    foreach (sort mysort keys %{$self->{dagwords}}) {
		$sentence .= " ${$self->{dagwords}}{$_}";
	    }
	    my $dag = join("\n","##DAG BEGIN $sentence",@dagedges,'##DAG END');
	    $label  = $self->{daglabel};
##	    $self->verbose("Handling sentence $label: $sentence");
	    $self->handle_sentence($label,'YES',$sentence,$dag);
##	    $self->verbose("Done Handling sentence $label: $sentence");
	    next;
	}

	next if /^#/;
	next if /^\s*$/;

	if ($dagmode) {
	    my $edge = $_;
	    chomp $edge;
	    push(@dagedges,$self->dag_decode($edge));
	    next;
	}

	## when we have a raw sentence as input
	my $status   = 'YES';
	my $sentence = $_;
	$label++;
	chomp $sentence;
	unless ($config->get('allparse')) {
	    if ($sentence =~ s/^>\s*//) {
		$status = 'SKIP';
	    } elsif ($sentence =~ s/^\*\s*//) {
		$status = 'NO';
	    }
	}
##	$self->verbose("Handling sentence $label: $sentence");
	$self->handle_sentence($label,$status,$sentence);
##	$self->verbose("Done Handling sentence $label: $sentence");
    }

    $self->{label} = $label;

    $io->close if ($io);
    
}

############################################
# Purpose    : handle a sentence (sentence or dag)
# Parameters : $label, $status, $sentence, $dag (decoded dag)
# Comments   : parse sentence and produce results in various formats
#
sub handle_sentence {
    my ($self, $label, $status, $sentence, $dag) = @_;
    #$self->verbose("handle_sentence $label");

    my $stats = $self->{stats};
    my $log   = $self->{log};
##    $self->{tmplog} = '';

    my $xsentence = $sentence;
    $xsentence =~ s/\{\{(.+?)\}\}//og;

    $self->{xsentence} = $xsentence ;
    $self->{status}    = $status ;
    $stats{tried}++;
    $self->{mark}      = ($status eq 'YES') ? ' ' : '*';
    $self->{label} = $label;

    my $cmd =  "$label $self->{parser} $sentence"; # parse sentence
    $cmd = "dag $cmd\n$dag" if (defined $dag); # parse dag
    $self->{sentence_init} = $cmd;
    $self->{tmplog} .= "$dag\n" if ($config->get('showdag') && defined $dag);
    $self->handle_sentence_aux($cmd);
  }

sub handle_sentence_aux {
  my ($self,$cmd) =@_;
  my $socket = $self->{socket};
  $self->{start_sentence} = [gettimeofday];
  ## $self->{watcher}->start; # from now, watch server output
  $self->send_cmd("$cmd");
}

############################################
# Purpose    : process server output (after 1st parsing)
#

sub build_yesno_handler {

  my $self = shift;
  my $verbose = $config->get('verbose');
  my $hostlabel = $self->{hostlabel};
  my $printhost_opt = $config->get('printhost');
  my $date_opt = $config->get('date');
  my $time_opt = $config->get('time');
  my $timeout = $config->get('timeout');
  my $status_opt = $config->get('status');

  return sub {
    my $label     = $self->{label};
    my $mark      = $self->{mark};
    my $xsentence = $self->{xsentence};
    my $sentence  = $self->{sentence};
    my $status    = $self->{status};
    my @output    = $self->get_output;
    
    ## no longer in sentnce_init mode
    delete $self->{sentence_init};
    
    my $restart = analyze_for_restart(\@output);
    
    my $answer    = 'NO'; # remains 'NO' when timeout
    $answer       = 'YES'     if (@output && ($output[0] =~ /Success/) );
    $answer       = 'PARTIAL' if (@output && ($output[0] =~ /Partial/) );
    
    if ($answer eq 'PARTIAL') {
      $self->{tmplog} .= <<EOF if ($verbose) ;
robust $label\t$mark $xsentence
EOF
  }
    elsif ($answer eq $status) {
      $self->{tmplog} .= <<EOF if ($verbose) ;
ok $label\t$mark $xsentence
EOF
    } else {
      $stats{failed}++;
      my $failed = $self->{failed};
      push(@$failed,{label=>$label,status=>$status,sentence=>$sentence});
      $self->{tmplog} .= <<EOF ;
**********************************************
fail $label\t$mark $xsentence
**********************************************
EOF
      
    }
  
    $self->tmplog(" <host> $hostlabel\n") if ($printhost_opt);
    $self->tmplog(strftime " <date> %F %H:%M:%S\n", localtime) if ($date_opt);
  
    if ($time_opt) {
      my $time;
      my $to = "";
      if ($restart) { # null or timeout
	$to = $restart;
	$time = $timeout;
      } else {
	$time = $self->{cmd_time} || 0;
      }
      $self->tmplog(" <time> $time$to\n");
      $stats{time} += $time;
      my $elapsed = tv_interval($self->{start_sentence});
      $self->tmplog(" <time1> $elapsed\n");
    }
    
    ## we can now restart if needed
    if ($restart) {
      my $elapsed = tv_interval($self->{start_sentence});
      $self->tmplog(" <timeproc> $elapsed\n");
      $self->flush_tmplog;
      $self->raise_connection_error("$restart (here1 $label yesno)");
      return;
    }
    
    ## if parsing doesn't fail, process handlers
    ##  $self->{watcher}->start;
    if ($answer ne 'NO') {
      $self->next_state;
    } else {
      $self->dag_finish;
    }
    
    report_stats() if ($status_opt);
  }
}

sub analyze_for_restart {
  my $output = shift;
  ## For the 2 following cases, we have to restart the connection
  ## but not until writing info about the sentence in the corpus log
  if (!@$output) { # maybe parserd is down !
    return " ** null output **";
  } elsif ($output->[0] eq 'timeout') {
    return " ** timeout **";
  } else {
    return '';
  }
}

sub get_handler {
  my ($self,$state) = @_;
  my $timeout = $config->get('timeout');

  if ($state eq 'stats') {
   return 
      sub {
	my $label     = $self->{label};
	my @output    = $self->get_output;

	if (my $restart = analyze_for_restart(\@output)) {
	  my $time = $timeout;
	  my $elapsed = tv_interval($self->{start_sentence});
	  $self->tmplog(" <timeamb> $time$restart\n",
			" <timeproc> $elapsed\n");
	  $self->flush_tmplog;
	  $self->raise_connection_error("$restart (here2: $label $state)");
	  return;
	}

	foreach my $info (@output) {
	  if ($info =~ s/^\s*Dependency stats:\s*//o) {
	    my ($clusters,$nodes,$edges) = (split(/\s+/,$info))[2,4,6];
	    $stats{clusters} += $clusters;
	    $stats{nodes} += $nodes;
	    $stats{edges} += $edges;
	    $self->tmplog(" <ambiguity> $info\n");
	  } else {
	    chomp $info;
	    $self->tmplog(" <ambiguity> ***pbm*** $info\n");
	  }
	}

	my $time = $self->{cmd_time};
	$self->tmplog(" <timeamb> $time\n");
	$self->next_state;
      };
 } elsif ($config->get('collbase') && $config->get('collsave')) {
   my $ext = $formats->{$state}{ext};
   return  
     sub {
       my $label     = $self->{label};
       my @output    = $self->get_output;

       if (my $restart = analyze_for_restart(\@output)) {
	 my $time = $timeout;
	 my $elapsed = tv_interval($self->{start_sentence});
	 $self->tmplog(" <time$state> $time$restart\n",
		       " <timeproc> $elapsed\n");
	 $self->flush_tmplog;
	 $self->raise_connection_error("$restart (here2: $label $state)");
	 return;
       }
	
       my $time = $self->{cmd_time};
       $self->tmplog(" <time$state> $time\n");
  
       ##	  $self->verbose("process $self->{output} $label");
       my $file =  $self->outfile($label,$ext);
       open(COLLS,">$file.bz2") || die "can't save in $file, stopped";
       print COLLS $output[0];
       close COLLS || die "can't close $file ($!)";

       $self->next_state;
     };
 } else  {
    return 
      sub {
	my $label     = $self->{label};
	my @output    = $self->get_output;

	if (my $restart = analyze_for_restart(\@output)) {
	  my $time = $timeout;
	  my $elapsed = tv_interval($self->{start_sentence});
	  $self->tmplog(" <time$state> $time$restart\n",
			" <timeproc> $elapsed\n");
	  $self->flush_tmplog;
	  $self->raise_connection_error("$restart (here2: $label $state)");
	  return;
	}
	  
	$self->tmplog("------------  START $state ---------\n",
		      @output,
		      "------------  END $state ---------\n");
	$self->next_state;
      };
  }
}

sub tmplog {
  my $self = shift;
  $self->{tmplog} .= $_ foreach (@_);
}

sub dag_finish {
  my $self = shift;
  my $log = $self->{log};
  ## flush log for current dag
  ##  $self->verbose("finish after $self->{output}");
  my $elapsed = tv_interval($self->{start_sentence});
  $self->tmplog(" <timeproc> $elapsed\n");
  $self->flush_tmplog;
  $self->{dispatcher}->next_dag($self)  if ($self->{dispatcher});
  if ($self->{final}) {
    $self->closing;
  }
}

sub set_init_state {
  my $self=shift;
  $self->{output} = 'yesno';
  my $info = $self->{states}{yesno};
  $self->{next} = $info->{next};
##  $self->{state_handler} = $info->{handler};
  $self->{watcher}->cb($info->{handler});
##  $self->verbose("init state to yesno");
}

sub next_state {
  my $self = shift;
  my $from = $self->{output};
  if (my $info = $self->{next}) {
    my $f = $self->{output} = $info->{format};
##    $self->verbose("switch state from $from to $f");
    $self->{next} = $self->{states}{$f}{next};
##    $self->{state_handler} = $self->{states}{$f}{handler};
    $self->{watcher}->cb($self->{states}{$f}{handler});
    $self->last_forest($info->{forest});
  } else {
    $self->dag_finish;
  }
}

sub report_stats {
    $config->report->finePrint('total',    0, $stats{tried});
    $config->report->finePrint('success', 10, $stats{tried} - $stats{failed});
    $config->report->finePrint('failed',  20, $stats{failed});
}

sub last_forest {
  my ($self, $type) = @_;
  ##    $self->verbose("last forest $self->{label} on $self->{host}");
  my $socket = $self->{socket};
  #     $self->{watcher}->stop;
  #     if ($config->get('disambiguate')) {
  # 	$self->send_directive_cmd("set disambiguate yes");
  #     }
  my $mode = $self->save_result ? 'bzip2' : 'data';
  $self->send_cmd("last forest $type $mode");
}

sub save_result {
  my $self = shift;
  my $kind = $self->{output};
  return exists $formats->{$kind}{ext};
}

sub mysort {
    my ($ea,$fa) = ($a =~ /E(\d+(?:\.\d+)?)F(\d+)/);
    my ($eb,$fb) = ($b =~ /E(\d+(?:\.\d+)?)F(\d+)/);
    return ($ea <=> $eb || $fa <=> $fb);
}

sub mysort2 {
    my ($ea) = ($a =~ /(\d+)$/);
    my ($eb) = ($b =~ /(\d+)$/);
    return ($ea <=> $eb);
}


############################################
# Purpose    : decode a dag edge and store words in a hash
# Parameters : $edge (a dag line)
# Comments   : example: takes the following edge input:
#              8     {<F id="E3F7">un</F>} "un"      9
#              returns:
#              7 {E3F7|un} "un" 8
sub dag_decode {
    my ($self, $edge) = @_;
    #print "dagdecode: edge: '$edge' daglabel '$self->{daglabel}'\n";
    my ($left,$comment,$token,$right) = ($edge =~ /^(\d+)\s+\{(.*?)\}\s+(\S+)\s+(\d+)/);
    my @comments = ();
    unless ($comment) {
	$self->verbose("$self->{hostlabel}: Bad dag format: $edge");
	return;
    }
    while ($comment =~ m{<F\s+id="(.+?)"\s*>\s*(.*?)\s*</F>}og) {
	my ($id,$w) = ($1,$2);
	$self->{dagwords}{$1} ||= $2;
	my $idw = "$id|$w";
	push(@comments,$idw) unless grep( $_ eq $idw,@comments);
    }
    $comment = join(' ',@comments);
    --$left;
    --$right;
    if ($left eq 0) {
	($self->{daglabel}) = ($comment =~ /^E(\d+(?:\.\d+)?)/);
    }
    #print  "$left {$comment} $token $right\n";
    return "$left {$comment} $token $right";
}

######################################################################
# all display handlers

## Specific handlers

sub display_handler {
    my $self = shift;
    my $display = $config->get('display');

    sub {
	my ($sentence, $label) = @_;
	my @lines = $self->last_forest($display);
	my $h = start ['recode','l1..u8'],
	'<pipe',\*DOT,,
	'|', ['dot','-Tgif',"-Glabel=$sentence"],
	'|', ['display'],
	'2>/dev/null'
	    or die "dot returned $?, stopped";
	print DOT @lines;
	close DOT;
	return $h;
    }
}

sub grammar_handler {
    my $self  = shift;

    sub {
	my @lines = $self->last_forest('html');
	my ($fh,$filename) = tempfile( DIR => '/tmp/',
				       SUFFIX => '.html',
	    );
	print $fh @lines;
	close $fh;
	my $h = start ['dillo',$filename],'>/dev/null','2>/dev/null';
	return $h;
    }
}

sub easyhtml_handler {
    my $self  = shift;

    sub {
	my @lines = $self->last_forest('easy');
	my ($fh,$filename) = tempfile( DIR => '/tmp/',
				       SUFFIX => '.html',
	    );
	close $fh;
	my $h = start 
	    #['xsltproc','-o',$filename,'@pkgdatadir@/easy2html.xsl','-'],'<pipe',\*EASY,
	    # tmp :
	    ['xsltproc','-o',$filename,'/home/pomerol/alpage/export-light/share/parserd/easy2html.xsl','-'],'<pipe',\*EASY,
	    ,'&',
	    ['dillo','-f','-l',$filename],'>/dev/null','2>/dev/null';
	print EASY @lines;
	close EASY;
	return $h;
    }
}

sub collection_handler {
    my $self  = shift;
    my @lines = $self->last_forest('easy');
    foreach (@lines) {
	next if (m{^<[/]?DOCUMENT});
	next if (m{<\?xml});
	print COLLECTION $_;
    }
}

sub collection_stop {
    my $self = shift;
    if ($self->{collection}) {
	print COLLECTION "</DOCUMENT>\n";
	close COLLECTION || die "can't close $self->{collection}, stopped";
    }
}

############################################
# Purpose    : Update collbase, colldir
# Parameters : corpus name, colldir
#
sub corpus_start {
    my $self = shift;
    $self->set_collbase(shift);
    $self->set_colldir(shift);
}

## why ?
sub collection_start {
    my $self    = shift;
    my $base    = shift;
    my $colldir = shift || $config->get('colldir');
    ## close old collection (if open) and start new one;
    $self->collection_stop;
    $self->{collbase} = $base;
    $config->colldir($colldir);
    my $coll = $self->{collection} = "$colldir/$base.easy.xml";
    open(COLLECTION,">$coll") || die "can't open $coll, stopped";
    autoflush COLLECTION;
    print COLLECTION <<EOF;
<?xml version="1.0"  encoding="latin1"?>
<DOCUMENT fichier="$base" xmlns:xlink="http://www.w3.org/1999/xlink">
EOF
}

sub outfile {
    my ($self, $label, $ext) = @_;
    my $l1 = floor($label / 10000);
    my $l2 = floor(($label - ($l1 * 10000)) / 1000);
    my $d = "$self->{colldir}/$l1/$l2";
    (-d $d) || system("mkdir -p $d");
    my $file = $d. "/".$self->{collbase}.".E$label.$ext";
    return $file;
}

sub set_collbase {
    my $self = shift;
    $self->{collbase} = shift;
}

sub set_colldir {
    my $self = shift;
    $self->{colldir} = shift;
}

sub all { $_ || return 0 for @_; 1 }

## The doc of IO::All mentions this method, which is actually not defined !
sub IO::All::appendf {
    my $self = shift;
    $self->assert_open('>>');
    $self->printf(@_);
}

sub final_stats {
    my $self = shift;
    my $log  = $self->{log};
    return unless ($self->{final} == 1);
    appendf $log "\nStats: %3u tried %3u failed %3u skipped %3.2f%% coverage %.2fs avtime\n", 
    $stats{tried}, 
    $stats{failed}, 
    $stats{skipped},
    100 * (1 - $stats{failed} / ($stats{tried} || 0.01)),
    $stats{time} / ($stats{tried} || 0.01);

    if ($config->get('stats')) {
	my $clusters = $stats{clusters};
	$clusters ||= 1;
	appendf $log "Stats:  %1.1f ambiguity\n",
	($stats{edges} - $clusters) / ($clusters || 0.01);
    }

    my $fail = $self->{failed};
    if (@$fail) {
	my $failed = @$fail;
	append $log <<EOF;

*** $failed tests failed !
EOF
    } elsif ($stats{tried} > 0) {
	append $log <<EOF;

All tests successful !
EOF

}
}

sub closing {
    my $self = shift;
    undef $self->{watcher};
    $self->send_cmd("quit") if ($self->{socket});
    $self->final_stats();
}

1;

=head1 NAME

dispatch.pl - to parse corpora on several hosts

=head1 SYNOPSIS

./dispatch.pl [options]

where the options are

=over 4

=item -allparse

=item -collbase=F<path>           basename of parsed sentences

=item -colldir=F<path>            [Default F<./results>]

=item -collsave                save a collection of sentences parses in colldir

=item -compress                compress output files with bz2

=item -date                    print date and time [Default 0]

=item -disambiguate|dis

=item -display|d dep           graphical dependency view

=item -dotdep

=item -easy                    XML Easy format

=item -easyhtml                HTML Easy format

=item -easyinput

=item -errfile=<file>          [Default errors]

=item -exclude=<corpus>        To exclude a corpus list [Default none]

=item -final                   [Default 1]

=item -forest                  XML shared derivations

=item -grammar                 HTML view of shared derivations

=item -host|h=<name>           [Default localhost]

=item -kill=<hours>            kill this script after <hours> (for grid'5000)

=item -ldep

=item -log_file=F<file>           [Default F</tmp/dispatch.log>]

=item -parser=<name>           run parser <name> [Default frmgtel]

=item -port|p=<number>         [Default 8999]

=item -robust                  test full parsing and, when failure, robust parsing

=item -rparser=<name>          [Default frmgtelr]

=item -setsid

=item -short

=item -showdag

=item -stats|s                 stats about ambiguity

=item -status

=item -tagger                  tagging projection of parsing

=item -input|in=<input>        input file or sentence

=item -time                    execution time

=item -timeout=<i>             [Default 200]

=item -verbose|v               [Default 1]

=item -xmldep                  XML shared dependencies

=back

=head1 DESCRIPTION

B<dispatch.pl> is a script to
run parsers with many options.

=head1 DOCUMENTATION

Please check the documentation at L<http://alpage.inria.fr/alpc.fr.html>.

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2004-2008, INRIA.

This program is free software; you can redistribute it and/or modify it under
the same terms as Perl itself.

=head1 AUTHOR

Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>

=cut
